#include <malloc.h>
#include "common.h"

/**
 *  max_int:
 *      Returns the maximum integer of two integers given.
 */
int max_int(int a, int b)
{
    return (a >= b) ? a : b;
}

/**
 *  min_int:
 *      Returns the minimum integer of two integers given.
 */
int min_int(int a, int b)
{
    return (a < b) ? a : b;
}

/**
 *  get_min_power_of_two_ge:
 *      Gets the minimum power of 2 that is greater than or equal to the
 *      specified number.
 *
 *  Remarks:
 *      For example, when the input is 1, the function returns 1; when the
 *      input is 2, the function returns 2; when the input is 3, the function
 *      returns 4.
 */
int get_min_power_of_two_ge(int n)
{
    int result;

    for (result = 1; result < n; result <<= 1) {
        ;
    }
    return result;
}

bool queue_between(int testee, int begin, int end, int queue_size)
{
    int linbeg; /* linear begin */
    int linend; /* linear end */
    int lintst; /* linear testee */

    linbeg = begin;
    linend = end;
    lintst = testee;
    if (linend < linbeg) {
        linend += queue_size;
    }
    if (lintst < linbeg) {
        lintst += queue_size;
    }
    return (bool)(linbeg <= lintst && lintst < linend);
}

/**
 *  re_place_last_in_order:
 *      When an array is in ascending order except for the last element, the
 *      function places the last element in the position that keeps the whole
 *      array in order. This can be seen as a step in Insertion Sort.
 *
 *  Parameters:
 *      base - the base address of the array
 *      num - number of elements in the array
 *      width - size in number of bytes of each element in the array
 *      compare - the comparison function. It accepts two void pointers
 *          pointing to two elements in the array, and returns a number above
 *          zero when the left element is greater than the right, or 0 if they
 *          equal, or below zero otherwise.
 */
void re_place_last_in_order(void *base, size_t num, size_t width, int (__cdecl *compare)(const void *, const void *))
{
    int i;
    void *buffer = alloca(width); /* temporary buffer for swapping */
    /* NOTE: alloca allocates on the stack and on failure generates a stack
     * overflow exception. */

    for (i = num - 1; i > 0; i--) {
        if (compare((void *)((char *)base + width * (i - 1)), (void *)
            ((char *)base + width * i)) <= 0)
        {
            break;
        }
        memcpy(buffer, (void *)((char *)base + width * i), width);
        memcpy((void *)((char *)base + width * i), (void *)((char *)
            base + width * (i - 1)), width);
        memcpy((void *)((char *)base + width * (i - 1)), buffer, width);
    }
}

