#include <assert.h>
#include <io.h>
#include <pdh.h>
#include <pdhmsg.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>
#include "common.h"
#include "resource.h"
#include "options.h"
#include "freememind.h"

/**
 *  freememind.c:
 *      The version with a window of the free memory performance monitor.
 */

#define WSHOWERRHDL(expression) if (!(expression)) {\
        MessageBox(hwnd, g_error_message, "Error - Free Memory Indicator",\
            MB_OK | MB_ICONSTOP);\
        clear_error_message();\
        DestroyWindow(hwnd);\
        goto lbl_finally;\
    }

#define WSHOWERRHDL2(expression) if (!(expression)) {\
        MessageBox(hwnd, g_error_message, "Error - Free Memory Indicator",\
            MB_OK | MB_ICONSTOP);\
        clear_error_message();\
        goto lbl_finally;\
    }

typedef BOOL (WINAPI *SetLayeredWindowAttributes_func)(HWND hwnd, COLORREF crKey, BYTE bAlpha, DWORD dwFlags);
typedef BOOL (WINAPI *GetLayeredWindowAttributes_func)(HWND hwnd, COLORREF *pcrKey, BYTE *pbAlpha, DWORD *pdwFlags);

BOOL CALLBACK MainDlgProc(HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam);
static void on_initdialog(HWND hwnd);
static void on_timer(HWND hwnd, WPARAM wparam, LPARAM lparam);
static void on_destroy(HWND hwnd);
static void init_maindlg_data(struct maindlg_data *data);
static void destroy_maindlg_data(struct maindlg_data *data);
static bool make_freemem_query(struct maindlg_data *data);
static bool do_query(struct maindlg_data *data, double *result, bool *diskbusy);
static bool make_counter_path(struct name *dest, PDH_COUNTER_PATH_ELEMENTS *src);
static bool get_counter_value(HCOUNTER hcounter, PDH_FMT_COUNTERVALUE *value);
static void load_dialog_icon(HWND hwnd);
static void auto_set_opacity(HWND hwnd, int alpha);
//static int auto_get_opacity(HWND hwnd);
static bool apply_options(HWND hwnd, struct maindlg_data *data);
static bool unapply_options(HWND hwnd, struct maindlg_data *data);
static void setwindow_topmost(HWND hwnd);
static HFONT simple_createfont(HDC hdc, const char *fontname, int ptsize);
static bool assign_logfont_to_options(struct option_pack *options, HDC hdc, LOGFONT *logfont);

char g_appName[] = "HelloWin";
struct maindlg_data *g_maindlg_data = NULL;
const char *FreeMemIni = "freememind.ini";
HINSTANCE g_hInstance = NULL;

int WINAPI WinMain(HINSTANCE hinst, HINSTANCE hprevinst, PSTR cmdline, int cmdshow)
{
    HWND hwnd = NULL;

    g_hInstance = hinst;
    g_maindlg_data = (struct maindlg_data *)malloc(sizeof(struct maindlg_data));
    if (g_maindlg_data == NULL) {
        add_error_message("Failed to malloc in WinMain.");
    }
    WSHOWERRHDL2(g_maindlg_data != NULL);
    init_maindlg_data(g_maindlg_data);
    DialogBox(hinst, MAKEINTRESOURCE(IDD_MAINDLG), NULL, MainDlgProc);
lbl_finally:
    if (g_maindlg_data != NULL) {
        destroy_maindlg_data(g_maindlg_data);
        free(g_maindlg_data);
    }
    return 0;
}

BOOL CALLBACK MainDlgProc(HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam)
{
    switch (message) {
    case WM_INITDIALOG:
        on_initdialog(hwnd);
        break;
    case WM_TIMER:
        on_timer(hwnd, wparam, lparam);
        break;
    case WM_LBUTTONDOWN:
        GetCursorPos(&g_maindlg_data->mousepos);
        ScreenToClient(hwnd, &g_maindlg_data->mousepos);
        SetCapture(hwnd);
        break;
    case WM_MOUSEMOVE:
        /* http://www.ragestorm.net/snippet?id=102 */
        if ((wparam & MK_LBUTTON) != 0 && GetCapture() == hwnd) {
            RECT rect;
            POINT p;
            GetWindowRect(hwnd, &rect); /* screen relative */
            /* I call this function instead of using lparam, because, sometimes
             * you might be out of the window area. */
            GetCursorPos(&p);
            ScreenToClient(hwnd, &p);
            /* This is the trick :)
             * Notice we use the relative position so the window position is
             * always(!) relative to the initial point we started the dragging.
             * What acutally moves the window is that we take into calculation
             * the current position of the mouse inside the client area. */
            MoveWindow(hwnd,
                rect.left - g_maindlg_data->mousepos.x + p.x,
                rect.top - g_maindlg_data->mousepos.y + p.y,
                rect.right - rect.left,
                rect.bottom - rect.top,
                TRUE);
        }
        break;
    case WM_LBUTTONUP:
        ReleaseCapture();
        break;
    case WM_CLOSE:
        EndDialog(hwnd, 0);
        break;
    case WM_DESTROY:
        on_destroy(hwnd);
        PostQuitMessage(0);
        break;
    case WM_ENDSESSION:
        DestroyWindow(hwnd);
        break;
    default:
        return FALSE;
    }
    return TRUE;
}

static void on_initdialog(HWND hwnd)
{
    FILE *fp = NULL;
    PDH_STATUS pdhstatus;
    UINT uret;/* holds UINT return values */

    load_dialog_icon(hwnd);
    setwindow_topmost(hwnd);
    if (access(FreeMemIni, 04) == 0) {
        fp = fopen(FreeMemIni, "r");
        if (fp == NULL) {
            add_error_message("Failed to open ini file in on_initdialog.\n");
        }
        WSHOWERRHDL(fp != NULL);
        WSHOWERRHDL(load_options(&g_maindlg_data->options, fp));
        WSHOWERRHDL(apply_options(hwnd, g_maindlg_data));
    }
    WSHOWERRHDL(make_freemem_query(g_maindlg_data));
    /* make an initial query to allow collecting of PhysicalDisk rate counters
     * */
    pdhstatus = PdhCollectQueryData(g_maindlg_data->hquery);
    if (pdhstatus != ERROR_SUCCESS) {
        add_error_message("Failed to collect the query data in on_initdialog.\n");
    }
    WSHOWERRHDL(pdhstatus == ERROR_SUCCESS);
    if (g_maindlg_data->options.poll_interval == 0) {
        g_maindlg_data->options.poll_interval = 5000;
    }
    uret = SetTimer(hwnd, 1, g_maindlg_data->options.poll_interval, NULL);
    if (uret == 0) {
        add_error_message("Failed to create the timer in on_initdialog.\n");
    }
    WSHOWERRHDL(uret != 0);
    g_maindlg_data->timeridx = 1;
    on_timer(hwnd, 1, 0);
lbl_finally:
    if (fp != NULL) {
        fclose(fp);
    }
}

static void on_timer(HWND hwnd, WPARAM wparam, LPARAM lparam)
{
    bool diskbusy;
    double value;
    /* `buffer' should be large enough if the value is from a year 2006
     * computer */
    char buffer[32];

    WSHOWERRHDL(do_query(g_maindlg_data, &value, &diskbusy));
    switch (g_maindlg_data->options.unit) {
    case MSU_BYTE:
        sprintf(buffer, "%.0lf", value);
        break;
    case MSU_KILOBYTE:
        sprintf(buffer, "%.1lfK", value / 1024.0);
        break;
    case MSU_MEGABYTE:
        sprintf(buffer, "%.1lfM", value / 1048576.0);
        break;
    default:
        assert(false);
        sprintf(buffer,"%.0lf", value);
        break;
    }
    if (diskbusy) {
        strcat(buffer, "*");
    }
    SetDlgItemText(hwnd, IDC_FREEMEMDISP, buffer);
lbl_finally:
    ;
}

static void on_destroy(HWND hwnd)
{
    FILE *fp = NULL;

    if (g_maindlg_data->dlg_destroyed) {
        return;
    }
    fp = fopen(FreeMemIni, "w");
    if (fp == NULL) {
        add_error_message("Failed to open the ini file in on_destroy.\n");
    }
    WSHOWERRHDL2(fp != NULL);
    WSHOWERRHDL2(unapply_options(hwnd, g_maindlg_data));
    WSHOWERRHDL2(save_options(&g_maindlg_data->options, fp));
lbl_finally:
    if (g_maindlg_data->timeridx != 0) {
        KillTimer(hwnd, 1);
        g_maindlg_data->timeridx = 0;
    }
    if (g_maindlg_data->hquery != INVALID_HANDLE_VALUE) {
        PdhCloseQuery(g_maindlg_data->hquery);
        g_maindlg_data->hquery = INVALID_HANDLE_VALUE;
    }
    g_maindlg_data->dlg_destroyed = true;
    if (fp != NULL) {
        fclose(fp);
    }
}

/**
 *  apply_options:
 *      Applies immediate options. Options that are not immediate such as
 *      polling interval will not be applied here.
 */
static bool apply_options(HWND hwnd, struct maindlg_data *data)
{
    bool retres = true;
    HFONT hnewfont = NULL;
    HDC hdc = NULL;
    struct option_pack *options;
    HWND hctrl;

    options = &data->options;
    hdc = GetDC(hwnd);
    RAISEERR(hdc != NULL,
        "Failed to get the dialog box device context in apply_options.\n");
    hnewfont = simple_createfont(hdc, options->fontname, options->fontsize);
    RAISEERR(hnewfont != NULL, "Failed to create the font in apply_options.\n");
    data->hfont = hnewfont;
    hnewfont = NULL;
    hctrl = GetDlgItem(hwnd, IDC_FREEMEMDISP);
    RAISEERR(hctrl != NULL, "apply_options: GetDlgItem IDC_FREEMEMDISP failed.\n");
    SendMessage(hctrl, WM_SETFONT, (WPARAM)data->hfont, MAKELPARAM(TRUE, 0));
    MoveWindow(hwnd, options->windowpos.x, options->windowpos.y,
        options->windowsize.cx, options->windowsize.cy, FALSE);
    auto_set_opacity(hwnd, (int)(options->opacity * 255.0 + 0.5));
lbl_finally:
    if (hnewfont != NULL) {
        DeleteObject(hnewfont);
    }
    if (hdc != NULL) {
        ReleaseDC(hwnd, hdc);
    }
    return retres;
}

/**
 *  unapply_options:
 *      Unapplies immediate options. Options that are not immediate such as
 *      polling interval will not be unapplied here.
 */
static bool unapply_options(HWND hwnd, struct maindlg_data *data)
{
    bool retres = true;
    RECT wndrect; /* window rect */
    HDC hdc = NULL;
    struct option_pack *options;
    HWND hctrl;

    options = &data->options;
    hdc = GetDC(hwnd);
    RAISEERR(hdc != NULL,
        "Failed to get the dialog box device context in apply_options.\n");
    if (data->hfont == NULL) {
        HFONT hfont;
        LOGFONT fontinfo;
        /* selects another font to get the dialog font */
        hctrl = GetDlgItem(hwnd, IDC_FREEMEMDISP);
        RAISEERR(hctrl != NULL, "Failed to get the HWND of the control in apply_options.\n");
        hfont = (HFONT)SendMessage(hctrl, WM_GETFONT, 0, 0);
        if (hfont == NULL) {
            hfont = GetStockObject(SYSTEM_FONT);
        }
        GetObject(hfont, sizeof(LOGFONT), &fontinfo);
        ERRHDL(assign_logfont_to_options(options, hdc, &fontinfo));
    } else {
        DeleteObject(data->hfont);
        data->hfont = NULL;
    }
    GetWindowRect(hwnd, &wndrect);
    options->windowpos.x = wndrect.left;
    options->windowpos.y = wndrect.top;
    options->windowsize.cx = wndrect.right - wndrect.left;
    options->windowsize.cy = wndrect.bottom - wndrect.top;
    /* no need to set `opacity' */
    /* no need to set `poll_interval' */
    /* no need to set `unit' */
lbl_finally:
    if (hdc != NULL) {
        ReleaseDC(hwnd, hdc);
    }
    return retres;
}

static void init_maindlg_data(struct maindlg_data *data)
{
    data->destroyed = false;
    data->dlg_destroyed = false;
    data->timeridx = 0;
    data->hquery = INVALID_HANDLE_VALUE;
    data->hmemcntr = INVALID_HANDLE_VALUE;
    data->hcachcntr = INVALID_HANDLE_VALUE;
    data->hdskcntr = INVALID_HANDLE_VALUE;
    data->hfont = NULL;
    data->mousepos.x = 0;
    data->mousepos.y = 0;
    init_option_pack(&data->options);
}

static void destroy_maindlg_data(struct maindlg_data *data)
{
    if (data == NULL) {
        return;
    }
    if (data->destroyed) {
        return;
    }
    assert(data->dlg_destroyed);
    assert(data->timeridx == 0);
    if (data->hquery != INVALID_HANDLE_VALUE) {
        PdhCloseQuery(data->hquery);
    }
    if (data->hfont != NULL) {
        DeleteObject(data->hfont);
    }
    destroy_option_pack(&data->options);
    data->destroyed = true;
}

static bool make_freemem_query(struct maindlg_data *data)
{
    bool retres = true;
    HQUERY hquery = INVALID_HANDLE_VALUE;
    PDH_STATUS pdhstatus;
    PDH_COUNTER_PATH_ELEMENTS ctpath_st; /* struct counter path */
    struct name ctpath_n; /* counter path name object */
    HCOUNTER hcounters[6];
    int hcidx; /* hcounters index */

    init_name(&ctpath_n);
    pdhstatus = PdhOpenQuery(NULL, 0, &hquery);
    RAISEERR(pdhstatus == ERROR_SUCCESS,
        "Failed to create the query in make_freemem_query.\n");
    hcidx = 0;
    /* create the counter path */
    ctpath_st.szMachineName = NULL;
    ctpath_st.szObjectName = "Memory";
    ctpath_st.szInstanceName = NULL;
    ctpath_st.szParentInstance = NULL;
    ctpath_st.dwInstanceIndex = (DWORD)-1;
    ctpath_st.szCounterName = "Available Bytes";
    ERRHDL(make_counter_path(&ctpath_n, &ctpath_st));
    /* add the counter */
    pdhstatus = PdhAddCounter(hquery, get_buffer_of_name(&ctpath_n),
        0, hcounters + hcidx);
    RAISEERR(pdhstatus == ERROR_SUCCESS,
        "Failed to add the first counter in make_freemem_query.\n");
    hcidx++;
    /* make another counter path and add */
    ctpath_st.szCounterName = "System Cache Resident Bytes";
    ERRHDL(make_counter_path(&ctpath_n, &ctpath_st));
    pdhstatus = PdhAddCounter(hquery, get_buffer_of_name(&ctpath_n),
        0, hcounters + hcidx);
    RAISEERR(pdhstatus == ERROR_SUCCESS,
        "Failed to add the second counter in make_freemem_query.\n");
    hcidx++;
    /* make another counter path and add */
    ctpath_st.szObjectName = "PhysicalDisk";
    ctpath_st.szInstanceName = "_Total";
    ctpath_st.szCounterName = "% Idle Time";
    ERRHDL(make_counter_path(&ctpath_n, &ctpath_st));
    pdhstatus = PdhAddCounter(hquery, get_buffer_of_name(&ctpath_n),
        0, hcounters + hcidx);
    if (pdhstatus != ERROR_SUCCESS) {
        /* maybe we are on NT 4 */
        hcounters[hcidx] = INVALID_HANDLE_VALUE;
    }
    /* save the data */
    data->hquery = hquery;
    hquery = INVALID_HANDLE_VALUE;
    hcidx = 0;
    data->hmemcntr = hcounters[hcidx++];
    data->hcachcntr = hcounters[hcidx++];
    data->hdskcntr = hcounters[hcidx++];
lbl_finally:
    if (hquery != INVALID_HANDLE_VALUE) {
        PdhCloseQuery(hquery);
    }
    destroy_name(&ctpath_n);
    return retres;
}

static bool do_query(struct maindlg_data *data, double *result, bool *diskbusy)
{
    bool retres = true;
    PDH_STATUS pdhstatus;
    PDH_FMT_COUNTERVALUE value;
    double summary;
    bool db;

    pdhstatus = PdhCollectQueryData(data->hquery);
    RAISEERR(pdhstatus == ERROR_SUCCESS,
        "Failed to collect the query data in do_query.\n");
    summary = 0.0;
    ERRHDL(get_counter_value(data->hmemcntr, &value));
    summary += value.doubleValue;
    ERRHDL(get_counter_value(data->hcachcntr, &value));
    summary += value.doubleValue;
    db = false;
    if (data->hdskcntr != INVALID_HANDLE_VALUE) {
        /* for the disk counter, use fault-tolerance */
        if (!get_counter_value(data->hdskcntr, &value)) {
            clear_error_message();
            db = true;
        } else {
            db = (bool)(value.doubleValue < 87.5);
        }
        /* collect again for the PhysicalDisk counter */
        pdhstatus = PdhCollectQueryData(data->hquery);
        RAISEERR(pdhstatus == ERROR_SUCCESS,
            "Failed to collect the query data in do_query.\n");
    }
    /* save results */
    *result = summary;
    *diskbusy = db;
lbl_finally:
    return retres;
}

static bool make_counter_path(struct name *dest, PDH_COUNTER_PATH_ELEMENTS *src)
{
    bool retres = true;
    DWORD size;
    PDH_STATUS pdhstatus;

    size = 0;
    pdhstatus = PdhMakeCounterPath(src, NULL, &size, 0);
    RAISEERR(pdhstatus != (PDH_STATUS)PDH_INVALID_ARGUMENT,
        "Invalid counter path argument encountered in make_counter_path.\n");
    ERRHDL(expand_name(dest, size));
    pdhstatus = PdhMakeCounterPath(src, get_buffer_of_name(
        dest), &size, 0);
    RAISEERR(pdhstatus == ERROR_SUCCESS,
        "Failed to make the counter path argument in make_counter_path.\n");
    /* resize the name structure */
    dest->length = strlen(get_buffer_of_name(dest));
lbl_finally:
    return retres;
}

static bool get_counter_value(HCOUNTER hcounter, PDH_FMT_COUNTERVALUE *value)
{
    bool retres = true;
    PDH_STATUS pdhstatus;

    pdhstatus = PdhGetFormattedCounterValue(hcounter, PDH_FMT_DOUBLE |
        PDH_FMT_NOSCALE | PDH_FMT_NOCAP100, NULL, value);
    RAISEERR(pdhstatus == ERROR_SUCCESS,
        "Failed to get the query data value in get_counter_value.\n");
lbl_finally:
    return retres;
}

static void load_dialog_icon(HWND hwnd)
{
    HICON hicon;

    hicon = LoadIcon(g_hInstance, MAKEINTRESOURCE(IDI_SMALL));
    if (hicon != NULL) {
        SendMessage(hwnd, WM_SETICON, (WPARAM)FALSE, (LPARAM)hicon);
    }
    hicon = LoadIcon(g_hInstance, MAKEINTRESOURCE(IDI_THISAPP));
    if (hicon != NULL) {
        SendMessage(hwnd, WM_SETICON, (WPARAM)TRUE, (LPARAM)hicon);
    }
}

static void auto_set_opacity(HWND hwnd, int alpha)
{
    long exstyle;
    HINSTANCE huserdll;
    SetLayeredWindowAttributes_func setattribs;

    if (alpha == 255) {
        return;
    }
    huserdll = LoadLibrary("user32.dll");
    if (huserdll == NULL) {
        return;
    }
    setattribs = (SetLayeredWindowAttributes_func)GetProcAddress(
        huserdll, "SetLayeredWindowAttributes");
    if (setattribs == NULL) {
        return;
    }
    exstyle = GetWindowLong(hwnd, GWL_EXSTYLE);
    exstyle |= WS_EX_LAYERED;
    SetWindowLong(hwnd, GWL_EXSTYLE, exstyle);
    setattribs(hwnd, RGB(0, 0, 0), (BYTE)alpha, LWA_ALPHA);
}

/**
 *  auto_get_opacity:
 *      Automatically gets the opacity (0~255).
 *
 *  Remarks:
 *      Setting translucency requires Windows 2000 but getting requires Windows
 *      XP, so do not use it.
 */
//static int auto_get_opacity(HWND hwnd)
//{
//    long exstyle;
//    HINSTANCE huserdll;
//    GetLayeredWindowAttributes_func getattribs;
//    BYTE alpha;
//    DWORD flag;
//
//    huserdll = LoadLibrary("user32.dll");
//    if (huserdll == NULL) {
//        return 255;
//    }
//    getattribs = (GetLayeredWindowAttributes_func)GetProcAddress(
//        huserdll, "GetLayeredWindowAttributes");
//    if (getattribs == NULL) {
//        MessageBox(hwnd, "Here", "Here", MB_OK);
//        return 255;
//    }
//    exstyle = GetWindowLong(hwnd, GWL_EXSTYLE);
//    if ((exstyle & WS_EX_LAYERED) == 0) {
//        return 255;
//    }
//    getattribs(hwnd, NULL, &alpha, &flag);
//    if (flag == LWA_ALPHA) {
//        return alpha;
//    }
//    return 255;
//}

static void setwindow_topmost(HWND hwnd)
{
    long exstyle;

    /* Try if it works on Windows 2000 and up: first activate the window
     * without setting the Z-order; then set the Z-order */
    SetWindowPos(hwnd, 0, 0, 0, 0, 0, SWP_NOZORDER | SWP_NOMOVE | SWP_NOSIZE);
    SetWindowPos(hwnd, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOACTIVATE |
        SWP_NOMOVE | SWP_NOSIZE);
    /* do not set WS_EX_TOPMOST style */
}

static HFONT simple_createfont(HDC hdc, const char *fontname, int ptsize)
{
    HFONT hfont;

    hfont = CreateFont(
        -MulDiv(ptsize, GetDeviceCaps(hdc, LOGPIXELSY), 72),
        0,
        0,
        0,
        FW_DONTCARE,
        FALSE,
        FALSE,
        FALSE,
        DEFAULT_CHARSET,
        OUT_DEFAULT_PRECIS,
        CLIP_DEFAULT_PRECIS,
        DEFAULT_QUALITY,
        DEFAULT_PITCH | FF_DONTCARE,
        fontname);
    return hfont;
}

static bool assign_logfont_to_options(struct option_pack *options, HDC hdc, LOGFONT *logfont)
{
    bool retres = true;
    char *newbuf = NULL;

    newbuf = (char *)malloc(sizeof(char) * get_min_power_of_two_ge(
        strlen(logfont->lfFaceName) + 1));
    RAISEERR(newbuf != NULL, "Failed to malloc in assign_logfont_to_options.");
    strcpy(newbuf, logfont->lfFaceName);
    assert(options->fontname == NULL);
    options->fontname = newbuf;
    newbuf = NULL;
    if (logfont->lfHeight < 0) {
        /* pixels to points */
        options->fontsize = -logfont->lfHeight * 72 / GetDeviceCaps(
            hdc, LOGPIXELSY);
    } else {
        /* don't know how to convert, use a default value */
        options->fontsize = 10;
    }
lbl_finally:
    if (newbuf != NULL) {
        free(newbuf);
    }
    return retres;
}

