#include <pdh.h>
#include <pdhmsg.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>
#include "common.h"

static bool do_things(void);
static bool make_counter_path(struct name *dest, PDH_COUNTER_PATH_ELEMENTS *src);
static bool get_counter_value(HCOUNTER hcounter, PDH_FMT_COUNTERVALUE *value);

#define MERRHDL(expression) if (!(expression)) {\
        retres = 1;\
        goto lbl_finally;\
    }


int main(void)
{
    int retres = 0;
    
    MERRHDL(do_things());
lbl_finally:
    if (retres == 1) {
        printf("Error!\n");
        show_and_clear_error();
    }
    return retres;
}

static bool do_things(void)
{
    bool retres = true;
    HQUERY hquery = INVALID_HANDLE_VALUE;
    PDH_STATUS pdhstatus;
    PDH_COUNTER_PATH_ELEMENTS cntrpath1;
    struct name cntrpath2;
    DWORD size;
    HCOUNTER hcounter1;
    HCOUNTER hcounter2;
    PDH_FMT_COUNTERVALUE value;
    double summary;

    init_name(&cntrpath2);
    pdhstatus = PdhOpenQuery(NULL, 0, &hquery);
    RAISEERR(pdhstatus == ERROR_SUCCESS,
        "Failed to create the query in do_things.\n");
    /* create the counter path */
    cntrpath1.szMachineName = NULL;
    cntrpath1.szObjectName = "Memory";
    cntrpath1.szInstanceName = NULL;
    cntrpath1.szParentInstance = NULL;
    cntrpath1.dwInstanceIndex = (DWORD)-1;
    cntrpath1.szCounterName = "System Cache Resident Bytes";
    ERRHDL(make_counter_path(&cntrpath2, &cntrpath1));
    /* add the counters */
    pdhstatus = PdhAddCounter(hquery, get_buffer_of_name(&cntrpath2),
        0, &hcounter1);
    RAISEERR(pdhstatus == ERROR_SUCCESS,
        "Failed to add the first counter in do_things.\n");
    cntrpath1.szCounterName = "Available Bytes";
    ERRHDL(make_counter_path(&cntrpath2, &cntrpath1));
    pdhstatus = PdhAddCounter(hquery, get_buffer_of_name(&cntrpath2),
        0, &hcounter2);
    RAISEERR(pdhstatus == ERROR_SUCCESS,
        "Failed to add the second counter in do_things.\n");
    /* query */
    pdhstatus = PdhCollectQueryData(hquery);
    RAISEERR(pdhstatus == ERROR_SUCCESS,
        "Failed to collect the query data in do_things.\n");
    summary = 0.0;
    ERRHDL(get_counter_value(hcounter1, &value));
    summary += value.doubleValue;
    ERRHDL(get_counter_value(hcounter2, &value));
    summary += value.doubleValue;
    printf("%lf\n", summary);
lbl_finally:
    if (hquery != INVALID_HANDLE_VALUE) {
        PdhCloseQuery(hquery);
    }
    destroy_name(&cntrpath2);
    return retres;
}

static bool make_counter_path(struct name *dest, PDH_COUNTER_PATH_ELEMENTS *src)
{
    bool retres = true;
    DWORD size;
    PDH_STATUS pdhstatus;

    size = 0;
    pdhstatus = PdhMakeCounterPath(src, NULL, &size, 0);
    RAISEERR(pdhstatus != (PDH_STATUS)PDH_INVALID_ARGUMENT,
        "Invalid counter path argument encountered in make_counter_path.\n");
    ERRHDL(expand_name(dest, size));
    pdhstatus = PdhMakeCounterPath(src, get_buffer_of_name(
        dest), &size, 0);
    RAISEERR(pdhstatus == ERROR_SUCCESS,
        "Failed to make the counter path argument in make_counter_path.\n");
    /* resize the name structure */
    dest->length = strlen(get_buffer_of_name(dest));
lbl_finally:
    return retres;
}

static bool get_counter_value(HCOUNTER hcounter, PDH_FMT_COUNTERVALUE *value)
{
    bool retres = true;
    PDH_STATUS pdhstatus;

    pdhstatus = PdhGetFormattedCounterValue(hcounter, PDH_FMT_DOUBLE |
        PDH_FMT_NOSCALE | PDH_FMT_NOCAP100, NULL, value);
    RAISEERR(pdhstatus == ERROR_SUCCESS,
        "Failed to get the query data value in get_counter_value.\n");
lbl_finally:
    return retres;
}
